#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Routes

Routes used by bank solution
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

#transport connection
CONNECTION = "amqp://guest:guest@localhost:5672/"
DEFAULT_EXCHANGE = ""

#client request ends
CLIENT_REQUEST_EXCHANGE = "lb.exchanges.Client"
CLIENT_REPLY_QUEUE = "lb.queues.Client"

#loan reply ends
LOAN_REPLY_EXCHANGE = "lb.exchanges.LoanBroker"
LOAN_REQUEST_QUEUE = "lb.queues.LoanBroker"

# Scoring gw ends
SCORING_GW_REQUEST_EXCHANGE = "lb.exchanges.ScoringGw"
SCORING_GW_REPLY_QUEUE = "lb.queues.ScoringGw"

#scoring ends
SCORING_REQUEST_QUEUE = "lb.queues.Scoring"

# Bank ends
BANK_GW_REPLY_QUEUE = "lb.queues.bankReply"
BANK_GW_CONTROL_QUEUE = "lb.control.bankGroups"
BANK_GW_CONTROL_EXCHANGE = "lb.control.bankGroups"

Routes = [
    # client loan relationships
    "direct://%s/%s" % (CLIENT_REQUEST_EXCHANGE, LOAN_REQUEST_QUEUE,),
    "direct://%s/%s" % (LOAN_REPLY_EXCHANGE, CLIENT_REPLY_QUEUE,),

    # loan scoring relationships
    "direct://%s/%s" % (SCORING_GW_REQUEST_EXCHANGE, SCORING_REQUEST_QUEUE,),
    "direct://%s/%s" % (DEFAULT_EXCHANGE, SCORING_GW_REPLY_QUEUE,),

    # Bank relationships
    "direct://%s/%s" % (BANK_GW_CONTROL_EXCHANGE, BANK_GW_CONTROL_QUEUE,),
    "direct://%s/%s" % (DEFAULT_EXCHANGE, BANK_GW_REPLY_QUEUE,),
]



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Bank

Bank instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import re
import random
import operator
import sleipnir.core

__all__ = ['BankGateway']

# Project requirements
from sleipnir.transport.connections import ConnectionFactory
from sleipnir.transport.patterns.endpoints import EndPoint
from sleipnir.transport.patterns.routers import EndPointRelay
from sleipnir.transport.patterns.aggregators import WrapperAggregate
from sleipnir.transport.patterns.aggregators import Aggregator, Header

# local submodule requirements

APP_ID = "app_id"
MESSAGE_ID = "message_id"
REPLY_TO = "reply_to"
CORRELATION_ID = "correlation_id"

from messages import BankQuoteReply


class Loan(object):

    __slots__ = ('_loans', '_nloan')

    # number of banks which propose loans
    expected = 0

    def __init__(self, expected):
        self._loans = []
        self._nloan = expected

    def add_message(self, message):
        self._loans.append(message)

    def is_complete(self, n_loans=None):
        n_loans = n_loans or self._nloan
        return len(self._loans) >= n_loans

    def get_message(self):
        best_loan = {
            str(BankQuoteReply.ERROR_CODE): 1,
            }
        best = 0.0

        for loan in self._loans:
            if loan[str(BankQuoteReply.ERROR_CODE)] == 0:
                interest_rate = loan[str(BankQuoteReply.INTEREST_RATE)]
                if best < interest_rate:
                    best, best_loan = interest_rate, loan
        return best_loan


class BankAggregate(WrapperAggregate):
    __slots__ = ()
    wrapper = Loan


class BankAggregator(EndPoint):

    producer = None

    def __init__(self, callback, request, error, connection=None):
        self._aggregator = Aggregator(
            Header(APP_ID),
            callback,
            BankAggregate)
        super(BankAggregator, self).__init__(request, None, error, connection)
        self._consumer.append_callbacks(self._aggregator.process)

    def __iter__(self):
        return iter(self._aggregator)

    def __getattr__(self, value):
        return getattr(self._aggregator, value)


class BankRelay(EndPointRelay):

    def get_eligible_banks(self, request):
        def parse(value):
            try:
                opr, value = re.match("([<>!=]{1,2})(.*)", str(value)).groups()
                value = int(value)
                # process
                return {
                    '>=': lambda x: operator.ge(x, value),
                    '=>': lambda x: operator.ge(x, value),
                    '=<': lambda x: operator.le(x, value),
                    '<=': lambda x: operator.le(x, value),
                    '<>': lambda x: operator.ne(x, value),
                    '==': lambda x: operator.eq(x, value),
                    '!=': lambda x: operator.ne(x, value),
                    '<':  lambda x: operator.lt(x, value),
                    '>':  lambda x: operator.gt(x, value),
                    }.get(opr, lambda x: False)
            except:
                return lambda x: False

        eligible = self._relay.keys()
        for bank, rules in self._relay.iteritems():
            rules = rules[0].iteritems()
            conditions = [(str(key), parse(val),) for key, val in rules]
            for key, func in conditions:
                if key not in request or not func(int(request[key])):
                    eligible.remove(bank)
                    break
        return eligible


class BankGw(object):

    class BankProcess(object):
        def __init__(self, callback, apply_id):
            self.callback = callback
            self.apply_id = apply_id

    def __init__(self, request, response, control, error, connection=None):
        self._apply_id = 0
        self._processes = {}
        self._response = BankAggregator(
            self._on_response, response, error, connection)
        self._rrequest = BankRelay(request, control, error, connection)

    def _on_response(self, response, message):
        if message:
            process = self._processes.pop(message.properties[APP_ID])
        else:
            # this is a hack to get bank gateway working when no
            # bank is present
            processes = self._processes.keys()
            [processes.remove(key) for key, _ in iter(self._response)]
            process = self._processes.pop(processes[0])
            assert len(processes) == 1
        process.callback(response, message)

    def manage_groups(self, name, banks):
        self._rrequest.route(name, banks)

    def best_quote(self, message, callback):
        banks = self._rrequest.get_eligible_banks(message)
        headers = {
            REPLY_TO: '@' + str(self._response.nodes.consumer.queues[0]),
            APP_ID:   str(random.randint(0, 10000000)),
            }

        # create a process and publish
        process = self.BankProcess(callback, headers[APP_ID])
        self._processes.setdefault(process.apply_id, process)
        self._rrequest.replay(message, headers, banks)

        # set completion value for aggregato
        aggregate = self._response.create(headers[APP_ID], len(banks))
        # enforces a response if no bank is online
        self._response.check(aggregate)

    def run(self, timeout=None):
        self._rrequest.run(timeout)
        self._response.run(timeout)


class ScoringGw(EndPoint):

    class ScoringProcess(object):
        def __init__(self, callback, apply_id):
            self.callback = callback
            self.apply_id = apply_id

    def __init__(self, request, response, error, connection=None):
        self._processes = {}
        super(ScoringGw, self).__init__(request, response, error, connection)
        self._consumer.append_callbacks(self.when_scoring_received)

    def get_scoring(self, request, callback):
        properties = {
            MESSAGE_ID: str(random.randint(0, 10000000)),
            REPLY_TO: self._consumer.queues[0],
            }
        # create a process
        process = self.ScoringProcess(callback, properties[MESSAGE_ID])
        self._processes.setdefault(process.apply_id, process)
        # prepare message and send
        message = self._producer.prepare(request, **properties)
        self._producer.publish(message)

    def when_scoring_received(self, response, message):
        correlation_id = message.properties[CORRELATION_ID]
        if correlation_id in self._processes:
            process = self._processes.pop(correlation_id)
            process.callback(response)
        else:
            # Send message to error queue
            raise NotImplementedError



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Scoring

Scoring instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import time
import random
import sleipnir.core

__all__ = ['Scoring']

# Project requirements
from sleipnir.transport.connections import ConnectionFactory
from sleipnir.transport.patterns.endpoints import Replier

# local submodule requirements
from routes import SCORING_REQUEST_QUEUE, CONNECTION
from messages import ScoringRequest, ScoringResponse


class Scoring(object):

    @property
    def score(self):
        return random.randint(300, 900)

    @property
    def history(self):
        return random.randint(1, 20)

    def process(self, request, properties):
        # compute result
        result = {
            str(ScoringResponse.SSN): request[str(ScoringRequest.SSN)],
            str(ScoringResponse.CREDIT_SCORE): self.score,
            str(ScoringResponse.HISTORY_LENGTH): self.history,
            }

        # sleep thread for msecs
        time.sleep(result[str(ScoringResponse.CREDIT_SCORE)] / 1000)
        # return loan result
        return (result, {})


if __name__ == '__main__':
    # Import here required modules for back standalone execution
    import sys

    # Verify parameters
    try:

        # create a Request reply pattern
        Replier(
            SCORING_REQUEST_QUEUE,
            None,
            CONNECTION, Scoring()).run()

        # notify how could be out
        sys.stdout.write("Press Ctrl-C to exit...\n")

        # start connection
        connection = ConnectionFactory.get_instance().create(CONNECTION)
        connection.loop.start()

    except KeyboardInterrupt:
        # close connection
        connection.close()



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Bank

Bank instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from random import randint

__all__ = ['LoanBroker']

# Project requirements
from sleipnir.core.decorators import cached
from sleipnir.core.utils import maybe_list, uuid
from sleipnir.transport.channels import ChannelEnd, Producer, Consumer
from sleipnir.transport.connections import ConnectionFactory
from sleipnir.transport.patterns.endpoints import Manager
from sleipnir.transport.patterns.translators import Translator

# local submodule requirements
import routes
from messages import *
from gateways import BankGw, ScoringGw


MESSAGE_ID = 'message_id'
CORRELATION_ID = 'correlation_id'


class LoanBrokerTranslator(Translator):
    def __init__(self):
        super(LoanBrokerTranslator, self).__init__()
        self.register(ScoringRequest, None)
        self.register(BankQuoteRequest, self._template_bank_quote_request)
        self.register(LoanQuoteReply, self._template_loan_quote_reply)

    def _template_bank_quote_request(self, result, klass, *values):
        loan_ssn = values[0][str(LoanQuoteRequest.SSN)]
        cred_ssn = values[1][str(ScoringResponse.SSN)]
        return result if (loan_ssn == cred_ssn) else None

    def _template_loan_quote_reply(self, result, klass, *values):
        if not result[str(LoanQuoteReply.QUOTE_ID)]:
            result[str(LoanQuoteReply.INTEREST_RATE)] = 0.0
            result[str(LoanQuoteReply.QUOTE_ID)] = "ERROR: No Qualifying Quote"
        return result


class LoanBrokerProcess(object):
    def __init__(self, manager, message):
        self._manager = manager
        self._message = message
        self._request = message.payload

    def _when_scoring_replied(self, message):
        translator, request = self._translator, self._request
        brq = translator(BankQuoteRequest).convert(request, message)
        brq and self._manager.banks.best_quote(brq, self._when_banks_replied)

    def _when_banks_replied(self, response, message):
        # prepare response
        props = {
            CORRELATION_ID: self._message.properties[MESSAGE_ID],
            MESSAGE_ID: uuid(),
            }
        translator, request = self._translator, self._request
        response = translator(LoanQuoteReply).convert(request, response)
        # now reply
        self._manager.reply(response, **props)
        self._manager.when_completed(props[CORRELATION_ID])

    def run(self, timeout=None):
        self._translator = translator = LoanBrokerTranslator()
        request = translator(ScoringRequest).convert(self._request)
        self._manager.scoring.get_scoring(request, self._when_scoring_replied)


class LoanBroker(Manager):

    # bank gateway
    bank = BankGw
    bank_opts = {
        'response': routes.BANK_GW_REPLY_QUEUE,
        'control':  routes.BANK_GW_CONTROL_QUEUE,
        'request':  None,
        'error':    None,
    }

    # scoring gateway
    cred = ScoringGw
    cred_opts = {
        'response': routes.SCORING_GW_REQUEST_EXCHANGE,
        'request':  routes.SCORING_GW_REPLY_QUEUE,
        'error':    None,
    }

    def __init__(self, bank=None, cred=None, endpoints=None, connection=None):
        super(LoanBroker, self).__init__(endpoints)
        self._bank = bank or self.bank(connection=connection, **self.bank_opts)
        self._cred = cred or self.cred(connection=connection, **self.cred_opts)

    def run(self, timeout=None):
        # start listening to clients
        self._bank.run(timeout)
        self._cred.run(timeout)
        # now listen to all consumers
        super(LoanBroker, self).run(timeout)

    def process(self, end, request, message):
        message_id = message.properties[MESSAGE_ID]
        lb_process = LoanBrokerProcess(self, message)
        self._processes.setdefault(message_id, lb_process).run()

    def reply(self, response, **properties):
        for producer in self._producers.itervalues():
            message = producer.prepare(response, **properties)
            producer.publish(message)

    def when_completed(self, message_id):
        del self._processes[message_id]

    @property
    def scoring(self):
        return self._cred

    @property
    def banks(self):
        return self._bank

if __name__ == '__main__':
    # Import here required modules for back standalone execution
    import sys

    try:
        # create manager endpoints
        endpoints = [
            ChannelEnd.create(
                routes.LOAN_REQUEST_QUEUE,
                Consumer,
                connection=routes.CONNECTION),
            ChannelEnd.create(
                routes.LOAN_REPLY_EXCHANGE,
                Producer,
                connection=routes.CONNECTION),
            ]
        # create manager and start
        LoanBroker(endpoints=endpoints, connection=routes.CONNECTION).run()

        # notify how could be out
        sys.stdout.write("Press Ctrl-C to exit...\n")

        # start connection
        connection = ConnectionFactory.get_instance().create(routes.CONNECTION)
        connection.loop.start()

    except KeyboardInterrupt:
        # close connection
        connection.close()



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Bank

Bank instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import random
import sleipnir.core

__all__ = ['Bank']

APP_ID = 'app_id'

# Project requirements
from sleipnir.transport.patterns.endpoints import Replier

from sleipnir.transport.connections import ConnectionFactory
from sleipnir.transport.channels import Producer
from sleipnir.transport.builder import Builder
from sleipnir.transport.enums import ChannelTargetState

# local submodule requirements
from routes import SCORING_REQUEST_QUEUE, BANK_GW_CONTROL_EXCHANGE
from routes import DEFAULT_EXCHANGE, CONNECTION
from messages import BankQuoteReply, BankQuoteRequest


class BankControl(Producer):
    def __init__(self, exchange, connection, bank):
        super(BankControl, self).__init__(exchange, connection)
        self.bank = bank

    def when_resumed(self, route, frame):
        limits, exchange = self.bank.limits, self.bank.queue
        self.publish(self.prepare(limits, reply_to="@" + exchange))

    def when_suspended(self, route, frame):
        self.publish(self.prepare(None, reply_to=self.bank.queue))


class Bank(object):

    # Banks rate
    rate_premium = 0.0

    def __init__(self, name, rate, maxloan, *args):
        self.name = name
        self.rate = float(rate)
        self.loan = int(maxloan)
        self.args = args
        # quote counter
        self.__counter = 0

    @property
    def limits(self):
        return {
            str(BankQuoteRequest.LOAN_AMOUNT): self.args[2],
            str(BankQuoteRequest.CREDIT_SCORE): self.args[0],
            str(BankQuoteRequest.HISTORY_LENGTH): self.args[1],
            }

    def iterest_rate(self, loanterm):
        if  loanterm > self.loan:
            return 0.0
        lterm = (loanterm / 12) / 10
        return self.rate + self.rate_premium + lterm + random.uniform(0, 1)

    def quote_id(self):
        qid = "%s-%0*d" % (self.name, 10, self.__counter)
        self.__counter += 1
        return qid

    def process(self, request, properties):
        tae = self.iterest_rate(request[str(BankQuoteRequest.LOAN_TERM)])
        # prepare result
        result = {
            str(BankQuoteReply.QUOTE_ID): self.quote_id(),
            str(BankQuoteReply.ERROR_CODE): 1 if not tae else 0,
            str(BankQuoteReply.INTEREST_RATE): tae,
            }
        # increment counter and return
        return (result, {APP_ID: properties[APP_ID]})


if __name__ == '__main__':
    # Import here required modules for back standalone execution
    import sys

    # Verify parameters
    if len(sys.argv) == 7:
        try:
            # Create bank service
            bank = Bank(*sys.argv[1:])
            bank.queue = "lb.queues." + bank.name

            # Setup control messages
            control = BankControl(BANK_GW_CONTROL_EXCHANGE, CONNECTION, bank)
            control.watch(state=ChannelTargetState.RESUMED, first=True)
            #FIXME: We can only add an state watch
            #control.watch(state=ChannelTargetState.SUSPENDED, first=True)

            # create a Request reply pattern
            replier = Replier(bank.queue, None, CONNECTION, bank)
            # wait till builder set up queu
            replier.nodes.consumer.watch(first=True)
            replier.run()

            # create route builder
            connection = ConnectionFactory.get_instance().create(CONNECTION)
            builder = connection.routes(Builder, force=True)
            builder("direct://%s/%s" % (DEFAULT_EXCHANGE, bank.queue))

            # notify how could be out
            sys.stdout.write("Press Ctrl-C to exit...\n")
            connection.loop.start()

        except KeyboardInterrupt:
            # close connection
            connection.close()
            connection.loop.start()
    else:
        sys.stderr.write(
            "Usage: " + sys.argv[0] + \
            " <name> <rate> <maxloan> <scoring>, <history> <amount>\n")



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Client

Client instance that process loan propositions
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import functools
import sleipnir.core

from random import randint
from time import sleep, time
from pprint import pprint

__all__ = ['Client']

# Project requirements
from sleipnir.core.utils import uuid
from sleipnir.transport.connections import ConnectionFactory
from sleipnir.transport.patterns.endpoints import Requestor

# local submodule requirements
from routes import CLIENT_REQUEST_EXCHANGE, CLIENT_REPLY_QUEUE, CONNECTION
from messages import LoanQuoteRequest

MESSAGE_ID = 'message_id'
CORRELATION_ID = 'correlation_id'
TIMESTAMP = 'timestamp'


class Client(object):
    def __init__(self, nmessages=100):
        self._num_messages = int(nmessages)
        self._buf_messages = {}

    def prepared(self, message):
        return message

    def request(self, message=None, properties={}):
        for count in xrange(0, self._num_messages):
            request = {
                str(LoanQuoteRequest.SSN): count,
                str(LoanQuoteRequest.LOAN_TERM): randint(0, 72) + 12,
                str(LoanQuoteRequest.LOAN_AMOUNT): randint(0, 20) * 5000 + 25000,
                }
            properties = {
                MESSAGE_ID: uuid(),
                TIMESTAMP: time(),
                }

            # return created message
            yield self._buf_messages.setdefault(count, (request, properties,))

    def published(self, request, message):
        self._buf_messages[message[3].message_id] = message
        ssn = request[str(LoanQuoteRequest.SSN)]
        sleep(0.100)

    def response(self, response, message):
        messageid = message.properties[CORRELATION_ID]
        timestamp = time() - self._buf_messages[messageid][3].timestamp
        del self._buf_messages[messageid]
        # show response
        print "Received: ", timestamp * 1000, " millisecs"
        pprint(response, indent=4)


if __name__ == '__main__':
    # Import here required modules for back standalone execution
    import sys

    # Verify parameters
    if len(sys.argv) == 2:
        try:
            # Create bank service
            client = Client(*sys.argv[1:])

            # Create a request reply pattern
            Requestor(
                CLIENT_REPLY_QUEUE,
                CLIENT_REQUEST_EXCHANGE,
                None,
                CONNECTION, client).run()

            # notify how could be out
            sys.stdout.write("Press Ctrl-C to exit...\n")

            # start connection
            connection = ConnectionFactory.get_instance().create(CONNECTION)
            connection.loop.start()

        except KeyboardInterrupt:
            # close connection
            connection.close()
    else:
        sys.stderr.write(
            "Usage: " + sys.argv[0] + " <num messages>\n")



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Enums

Common structs used by bank solution
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import sleipnir.core

__all__ = [
    "BankQuoteReply",
    "BankQuoteRequest",
    "LoanQuoteReply",
    "LoanQuoteRequest",
    "ScoringRequest",
    "ScoringResponse",
    ]


class BankQuoteRequest(enum):
    SSN,            \
    CREDIT_SCORE,   \
    HISTORY_LENGTH, \
    LOAN_TERM,      \
    LOAN_AMOUNT = xrange(0, 5)


class BankQuoteReply(enum):
    INTEREST_RATE,  \
    QUOTE_ID,       \
    ERROR_CODE = xrange(0, 3)


class CreditBureauReply(enum):
    SSN, = xrange(0, 1)


class LoanQuoteRequest(enum):
    SSN,       \
    LOAN_TERM, \
    LOAN_AMOUNT = xrange(0, 3)


class LoanQuoteReply(enum):
    SSN,         \
    QUOTE_ID,    \
    LOAN_AMOUNT, \
    INTEREST_RATE = xrange(0, 4)


class ScoringRequest(enum):
    SSN, = xrange(0, 1)


class ScoringResponse(enum):
    SSN,          \
    CREDIT_SCORE, \
    HISTORY_LENGTH = xrange(0, 3)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Setup

Setup routes
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
import pika.log

# Project requirements
from sleipnir.transport.builder import Builder
from sleipnir.transport.channels import Producer
from sleipnir.transport.connections import ConnectionFactory

# local submodule requirements
import routes


class BankBuilder(Builder):
    # Number of routes already builded
    builded = 0

    def when_resumed(self, route, frame):
        self.builded += 1
        if len(routes.Routes) == self.builded:
            print "routes established. Closing now..."
            self._connection.close()
            self._connection.loop.start()

if __name__ == '__main__':
    # set level error
    pika.log.setup(level=pika.log.INFO)

    # create a connection
    connect = ConnectionFactory.get_instance().create(routes.CONNECTION)
    # create a builder and routes
    builder = connect.routes(BankBuilder, force=True)
    [builder(route) for route in routes.Routes]

    # create now!
    connect.loop.start()
